#ifndef _WIN32
#error This is only for Windows targets.
#endif

#include "internal.h"
#include "midiseq.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <ctype.h>

#ifndef MOD_WAVETABLE
#define MOD_WAVETABLE 6
#define MOD_SWSYNTH 7
#endif

#define STREAMBUFLEN 4096	// *4 = 16K of event data per quarter note (!)

static midiseq *midi = NULL;
static char paused = 0;
static HMIDISTRM streamhnd = NULL;
static UINT mididevicenum = 0;

static MIDIHDR headers[2];
static DWORD bufdata[2][STREAMBUFLEN], buffullness[2];
static int buffersplaying = 0;

static HANDLE hSemBuffers = NULL;
static HANDLE hEvtExit = NULL;
static HANDLE hThreadBuffer = NULL;

static unsigned long evtdelta;
static unsigned char evtcommand = 0;
static unsigned char const *evtdata;
static unsigned long evtlength;

static int bufferdata(int num)
{
	MMRESULT result;
	
	// buffers some messages and queues them
	unsigned long tickamt;
	int i=0,j,k;
	
	if (!midi) return 1;

	buffullness[num] = 0;

	if (!evtcommand) {
		if (midi->getevent(&evtdelta, &evtcommand, &evtdata, &evtlength)) {
			evtcommand = 0;
			return 1;	// error or eof
		}
	}

	tickamt = midi->getppqn();
	
	if (paused) {
		// queue a nop for tickamt
		bufdata[num][ buffullness[num]++ ] = tickamt;
		bufdata[num][ buffullness[num]++ ] = 0;
		bufdata[num][ buffullness[num]++ ] = (MEVT_NOP<<24);
		goto queueandret;
	}

	if (evtdelta > tickamt) {
		evtdelta -= tickamt;

		// queue a nop for tickamt
		bufdata[num][ buffullness[num]++ ] = tickamt;
		bufdata[num][ buffullness[num]++ ] = 0;
		bufdata[num][ buffullness[num]++ ] = (MEVT_NOP<<24);
		goto queueandret;
	}

	do {
		// queue the event. tempo events get inserted as MEVT_TEMPO
		tickamt -= evtdelta;

		bufdata[num][ buffullness[num]++ ] = evtdelta;
		bufdata[num][ buffullness[num]++ ] = 0;
		if (evtcommand == 0xFF && evtdata[0] == 0x51) {
			bufdata[num][ buffullness[num]++ ] = (MEVT_TEMPO<<24) | ((DWORD)evtdata[2]<<16) |
			                                     ((DWORD)evtdata[3]<<8) | ((DWORD)evtdata[4]);
		} else if (evtlength < 3) {
			bufdata[num][ buffullness[num] ] = MEVT_F_SHORT | evtcommand;
			switch (evtlength) {
				case 2: bufdata[num][ buffullness[num] ] |= (DWORD)evtdata[1] << 16;
				case 1: bufdata[num][ buffullness[num] ] |= (DWORD)evtdata[0] << 8;
				default: break;
			}
			buffullness[num]++;
		} else {
			bufdata[num][ buffullness[num]++ ] = MEVT_F_LONG | (1+evtlength);
			bufdata[num][ buffullness[num] ] = evtcommand;
			for(j=1,k=0;k<evtlength;k++) {
				bufdata[num][ buffullness[num] ] |= ((DWORD)evtdata[k] << (j*8));
				if (++j==4) {
					buffullness[num]++;
					j=0;
					bufdata[num][ buffullness[num] ] = 0;
				}
			}
			if (j>0) buffullness[num]++;	// ended on a partial?
		}

		i = midi->getevent(&evtdelta, &evtcommand, &evtdata, &evtlength);
		if (i>0) {
			evtcommand = 0;
			break;	// end of song
		} else if (i<0) {
			evtcommand = 0;
			return 1;	// error or eof
		}
	} while (evtdelta <= tickamt && STREAMBUFLEN-buffullness[num] > (evtlength+sizeof(MIDIEVENT)+3)/4);

queueandret:
	// queue buffer
	headers[num].lpData = (LPSTR)&bufdata[num];
	headers[num].dwBufferLength = headers[num].dwBytesRecorded = buffullness[num]*4;
	headers[num].dwUser = num;
	headers[num].dwFlags = 0;
	result = midiOutPrepareHeader((HMIDIOUT)streamhnd, &headers[num], sizeof(MIDIHDR));
	if (result != MMSYSERR_NOERROR) return -1;
	result = midiStreamOut(streamhnd, &headers[num], sizeof(MIDIHDR));
	if (result != MMSYSERR_NOERROR) return -1;
	buffersplaying++;

	return i;
}

// is it safe to buffer in the callback?
static void CALLBACK midiCallback(HMIDIOUT handle, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	LPMIDIHDR hdr;

	if (uMsg != MOM_DONE) return;

	hdr = (LPMIDIHDR)dwParam1;
	midiOutUnprepareHeader(handle,hdr,sizeof(MIDIHDR));
	buffersplaying--;
	buffullness[hdr->dwUser] = 0;
	ReleaseSemaphore(hSemBuffers, 1, NULL);
	//bufferdata(hdr->dwUser);	// dwUser contains the buffer's index in headers[]
}

static DWORD WINAPI buffThread(LPVOID lpParam)
{
	DWORD waitret;
	HANDLE waithnds[2];
	int i,done=0;

	waithnds[0] = hSemBuffers;
	waithnds[1] = hEvtExit;

	while (!done) {
		waitret = WaitForMultipleObjects(2,waithnds,FALSE,INFINITE);
		if (waitret == WAIT_OBJECT_0) {
			for (i=0;i<2;i++) {
				if (buffullness[i] > 0) continue;
				if (bufferdata(i)) done = 1;
				break;
			}
		} else if (waitret == WAIT_OBJECT_0+1) {
			done = 1;
		} else {
			break;
		}
	}

	return 0;
}

jfauderr jfaudmidi_stop(void)
{
	int i;
	
	if (streamhnd) {
		midiStreamStop(streamhnd);

		for (i=0;i<2;i++)
			if (headers[i].dwFlags & MHDR_PREPARED)
				midiOutUnprepareHeader((HMIDIOUT)streamhnd, &headers[i], sizeof(MIDIHDR));

		midiStreamClose(streamhnd);
		streamhnd = NULL;
	}

	if (hSemBuffers) {
		CloseHandle(hSemBuffers);
		hSemBuffers = NULL;
	}

	if (hEvtExit) {
		SetEvent(hEvtExit);	// causes the buffer thread to die
		CloseHandle(hEvtExit);
		hEvtExit = NULL;
	}

	if (hThreadBuffer) {
		CloseHandle(hThreadBuffer);
		hThreadBuffer = NULL;
	}

	return jfauderr_ok;
}

jfauderr jfaudmidi_play(midiseq *file, int pause)
{
	MMRESULT result;
	MIDIPROPTIMEDIV tdiv;
	int i;

	jfaudmidi_stop();

	memset(headers, 0, sizeof(headers));
	memset(bufdata, 0, sizeof(bufdata));
	memset(buffullness, 0, sizeof(buffullness));

	midi = file;
	paused = pause;
	buffersplaying = 0;

	// opens the midi device, send a reset
	result = midiStreamOpen(&streamhnd, &mididevicenum, 1, (DWORD)midiCallback, 0, CALLBACK_FUNCTION);
		// dwInstance would be the pointer to the stream info struct
	if (result != MMSYSERR_NOERROR) {
		return jfauderr_genericerror;
	}
	
	// set timebase from file
	tdiv.cbStruct = sizeof(MIDIPROPTIMEDIV);
	tdiv.dwTimeDiv = file->getppqn();
	result = midiStreamProperty(streamhnd, (LPBYTE)&tdiv, MIDIPROP_SET|MIDIPROP_TIMEDIV);
	if (result != MMSYSERR_NOERROR) {
		jfaudmidi_stop();
		return jfauderr_genericerror;
	}
	
	// start the buffering thread
	hSemBuffers = CreateSemaphore(NULL, 2, 2, NULL);
	hEvtExit = CreateEvent(NULL, FALSE, FALSE, NULL);
	hThreadBuffer = CreateThread(NULL, 0, buffThread, NULL, 0, (LPDWORD)&i);
	if (!hSemBuffers || !hEvtExit || !hThreadBuffer) {
		jfaudmidi_stop();
		return jfauderr_genericerror;
	}

	// start playing
	result = midiStreamRestart(streamhnd);
	if (result != MMSYSERR_NOERROR) {
		jfaudmidi_stop();
		return jfauderr_genericerror;
	}

	return jfauderr_ok;
}

static jfauderr isrunning(void)
{
	DWORD st = 0;
	if (!hThreadBuffer) return jfauderr_notinited;
	if (!GetExitCodeThread(hThreadBuffer, &st)) return jfauderr_genericerror;
	if (st==STILL_ACTIVE) return jfauderr_ok;
	return (buffersplaying>0)?jfauderr_ok:jfauderr_genericerror;
}

jfauderr jfaudmidi_setplaymode(jfaudplaymode mode)
{
	if (isrunning() != jfauderr_ok) return jfauderr_notplaying;
	else if (mode == jfaud_stopped) {
		jfaudmidi_stop();
		return jfauderr_ok;
	} else if (mode == jfaud_paused) {
		paused = 1;
		return jfauderr_ok;
	} else if (mode == jfaud_playing) {
		paused = 0;
		return jfauderr_ok;
	}
	return jfauderr_genericerror;
}

jfauderr jfaudmidi_getplaymode(jfaudplaymode *mode)
{
	if (isrunning() != jfauderr_ok) *mode = jfaud_stopped;
	else if (paused) *mode = jfaud_paused;
	else *mode = jfaud_playing;
	return jfauderr_ok;
}

jfauderr jfaudmidi_setloop(int onf)
{
	if (!midi) return jfauderr_genericerror;
	midi->setloop(onf);
	return jfauderr_ok;
}

int midicardtype = Card_GeneralMIDI;

jfauderr jfaudmidi_init(void)
{
	struct jfaud_init_params *params;
	int numdevs, i, j, matchdev = -1, matchchars = -1;
	MIDIOUTCAPS caps, matchcaps;
	
	// figure out which midi device to select, and what type of card it resembles
	
	params = jfaud_getparams();

	numdevs = midiOutGetNumDevs();
	if (!numdevs) {
		params->call_logmsg("No MIDI devices available.");
		return jfauderr_nomidi;
	}

	params->call_logmsg("Available MIDI devices:");
	if (!params->mididevice[0]) matchdev=0;
	for (i=0; i<numdevs; i++) {
		if (midiOutGetDevCaps(i,&caps,sizeof(MIDIOUTCAPS)) != MMSYSERR_NOERROR) continue;

		for (j=0; caps.szPname[j] && params->mididevice[j] &&
		          toupper(caps.szPname[j]) == toupper(params->mididevice[j]);
			j++) ;
		if (j > matchchars) matchchars = j, matchdev = i, memcpy(&matchcaps,&caps, sizeof(MIDIOUTCAPS));
		
		params->call_logmsg(jsprintf("  - %s",caps.szPname));
	}
	params->call_logmsg(jsprintf("Selected device: %s",matchcaps.szPname));
	switch (matchcaps.wTechnology) {
		case MOD_MIDIPORT:
			params->call_logmsg("Device is a MIDI hardware port.");
			midicardtype = Card_GeneralMIDI;
			break;
		case MOD_SYNTH:
			params->call_logmsg("Device is a generic synthesiser.");
			midicardtype = Card_GeneralMIDI;
			break;
		case MOD_SQSYNTH:
			params->call_logmsg("Device is a square-wave synthesiser.");
			midicardtype = Card_OPL2;
			break;
		case MOD_FMSYNTH:
			params->call_logmsg("Device is an FM synthesiser.");
			midicardtype = Card_OPL2;
			break;
		case MOD_WAVETABLE:
			params->call_logmsg("Device is a wavetable synthesiser.");
			midicardtype = Card_GeneralMIDI;
			break;
		case MOD_SWSYNTH:
			params->call_logmsg("Device is a software synthesiser.");
			midicardtype = Card_GeneralMIDI;
			break;
		default:
			params->call_logmsg("Device is some unusual type of synthesiser.");
			midicardtype = Card_GeneralMIDI;
			break;
	}
	mididevicenum = matchdev;

	return jfauderr_ok;
}

jfauderr jfaudmidi_uninit(void)
{
	jfaudmidi_stop();
	return jfauderr_ok;
}

