#ifndef __jfaud_h__
#define __jfaud_h__

#ifdef __cplusplus
extern "C" {
#endif

struct jfaud_init_params {
	char *wavedevice;			// OpenAL audio device string
		// NULL = no waveform audio
		// empty string = default
		// device name
	char *mididevice;			// MIDI device string
		// NULL = no MIDI
		// empty string = midi mapper
		// (win32) device name
		// "*adlibemu" = Ken's adlib emulator (requires a pcm device)
	char *cdadevice;			// CD-A device string
		// NULL = no CD-A
		// empty string = first CDROM drive
		// (win32) drive letter
		// (*nix) cdrom device

	unsigned int maxvoices;			// maximum number of sounds to play simultaneously
	unsigned int samplerate;		// sampling frequency to try and use

	int (*call_open)(const char *fn);	// file opening callback
	int (*call_close)(int h);		// file closing callback
	int (*call_read)(int h, void *buf, int len);	// file reading callback
	int (*call_seek)(int h, int off, int whence);	// file seeking callback
	int (*call_tell)(int h);		// file telling callback

	void (*call_logmsg)(const char *str);	// log message output callback
};

typedef enum {
	jfauderr_ok = 0,
	
	jfauderr_genericerror = -1,
	jfauderr_outofmemory = -2,
	
	jfauderr_notinited = 1,
	jfauderr_alreadyinited,
	
	jfauderr_openerror,
	
	jfauderr_unrecognised,
	jfauderr_wrongtype,
	jfauderr_nowave,
	jfauderr_nomidi,
	jfauderr_nocda,

	jfauderr_notplaying,
	jfauderr_notloaded,
	jfauderr_notprepared,

	jfauderr_badhandle
} jfauderr;

const char *jfauderr_getstring(jfauderr err);

typedef enum {
	jfaud_stopped = 0,
	jfaud_paused = 1,
	jfaud_playing = 2
} jfaudplaymode;

// This character should be used for delimiting the subfilename from the
// base filename in files loaded from containers.
#define JFAUD_SUBFN_CHAR '|'
#define JFAUD_SUBFN_STR "|"

jfauderr jfaud_init(struct jfaud_init_params *params);
jfauderr jfaud_uninit(void);
jfauderr jfaud_isinited(void);
jfauderr jfaud_update(void);

jfauderr jfaud_playmusic(const char *file, float volume, char loop, jfaudplaymode playmode);
jfauderr jfaud_setmusicplaymode(jfaudplaymode playmode);
jfauderr jfaud_getmusicplaymode(jfaudplaymode *playmode);

jfauderr jfaud_playsound(int *handle, const char *file,
		float volume, float pitch, char loop,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel, char threed, int priority, jfaudplaymode playmode);
jfauderr jfaud_movesound(int handle, float posx, float posy, float posz,
		float velx, float vely, float velz, char posrel);
jfauderr jfaud_setpos(float posx, float posy, float posz,
		float velx, float vely, float velz,
		float atx, float aty, float atz,
		float upx, float upy, float upz);
jfauderr jfaud_setsoundplaymode(int handle, jfaudplaymode playmode);
jfauderr jfaud_getsoundplaymode(int handle, jfaudplaymode *playmode);
jfauderr jfaud_simulateaudiolib(int onf);

#ifdef __cplusplus
}
#endif

#endif
