#include "internal.h"

#include <stdlib.h>


jfauderr jfaudbuffer_alloc(jfaud_buffer *buf, int samplerate, short channels, short bitspersample, int numsamples)
{
	buf->samplerate = samplerate;
	buf->channels = channels;
	buf->bitspersample = bitspersample;
	buf->albuf = 0;
	buf->data = malloc(channels * (bitspersample / 8) * numsamples);
	if (!buf->data) return jfauderr_outofmemory;

	buf->numsamples = numsamples;

	return jfauderr_ok;
}

jfauderr jfaudbuffer_enlarge(jfaud_buffer *buf, int numsamples)
{
	void *b;
	b = realloc(buf->data, buf->channels * (buf->bitspersample / 8) * numsamples);
	if (!b) return jfauderr_outofmemory;

	buf->data = b;
	buf->numsamples = numsamples;

	return jfauderr_ok;
}

jfauderr jfaudbuffer_free(jfaud_buffer *buf)
{
	if (buf->data) free(buf->data);
	buf->data = NULL;
	buf->numsamples = 0;

	return jfauderr_ok;
}


static struct jfaud_init_params *params;

#define STALEAGE 256
jfaud_bufferlink *managed_head = NULL, *managed_tail = NULL;


jfauderr jfaudbuffer_init(void)
{
	params = jfaud_getparams();
	managed_head = managed_tail = NULL;
	return jfauderr_ok;
}

jfauderr jfaudbuffer_uninit(void)
{
	jfaud_bufferlink *lp;

	while (managed_head) {
		lp = managed_head->next;
		jfaudopenal_releasebuffer(&managed_head->buffer);
		jfaudbuffer_free(&managed_head->buffer);
		free(managed_head);
		managed_head = lp;
	}
	managed_head = managed_tail = NULL;
	return jfauderr_ok;
}

jfauderr jfaudbuffer_fetch(jfaudident ident, jfaud_buffer **buf)
{
	jfaud_bufferlink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident == ident) {
			*buf = &lp->buffer;
			lp->ref++;
			debuglog(params->call_logmsg,jsprintf("buffer: Fetched ident=%u",ident));
			return jfauderr_ok;
		}
		lp = lp->next;
	}

	*buf = NULL;
	return jfauderr_ok;
}

jfauderr jfaudbuffer_isavailable(jfaudident ident)
{
	jfaud_bufferlink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident == ident) return jfauderr_ok;
		lp = lp->next;
	}

	return jfauderr_notloaded;
}

jfauderr jfaudbuffer_release(jfaudident ident)
{
	jfaud_bufferlink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident == ident) {
			if (lp->ref > 0) {
				lp->ref--;
				lp->age = 0;
				debuglog(params->call_logmsg,jsprintf("buffer: Released ident=%u",ident));
			}

			return jfauderr_ok;
		}
		lp = lp->next;
	}

	return jfauderr_genericerror;
}

jfauderr jfaudbuffer_manage(jfaudident ident, jfaud_buffer *buf)
{
	// allocate an entry on the management list
	jfaud_bufferlink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident == ident) break;
		lp = lp->next;
	}

	if (lp) {
		// replace an entry
		jfaudbuffer_free(&lp->buffer);
	} else {
		lp = (jfaud_bufferlink*)calloc(1, sizeof(jfaud_bufferlink));
		if (!lp) return jfauderr_outofmemory;
		
		if (!managed_tail) {
			managed_head = managed_tail = lp;
		} else {
			managed_tail->next = lp;
			lp->prev = managed_tail;
			managed_tail = lp;
		}

		lp->ref = 0;
		lp->ident = ident;
	}

	memcpy(&lp->buffer, buf, sizeof(jfaud_buffer));
	debuglog(params->call_logmsg,jsprintf("buffer: Added ident=%u for management",ident));

	return jfauderr_ok;
}

static jfaud_bufferlink *updateptr = NULL;
jfauderr jfaudbuffer_update(void)
{
	jfaud_bufferlink *lpb;

	if (!updateptr) updateptr = managed_head;
	if (!updateptr) return jfauderr_ok;

	if (updateptr->ref > 0) {
		updateptr = updateptr->next;
	} else if (updateptr->age < STALEAGE) {
		updateptr->age++;
	} else {
		lpb = updateptr->next;

		if (updateptr->prev) updateptr->prev->next = updateptr->next;
		else {
			managed_head = updateptr->next;
			if (managed_head) managed_head->prev = NULL;
		}
		if (updateptr->next) updateptr->next->prev = updateptr->prev;
		else {
			managed_tail = updateptr->prev;
			if (managed_tail) managed_tail->next = NULL;
		}
			
		jfaudbuffer_free(&updateptr->buffer);
		debuglog(params->call_logmsg,jsprintf("buffer: Removed stale ident=%u",updateptr->ident));
		free(updateptr);

		updateptr = lpb;
	}

	return jfauderr_ok;
}
