/*
 * This is probably the most hate-ridden piece of sin I've written
 * in a long time. Inspiration and direction for the decoding drawn
 * from audio_mad.c from the gmerlin_avdecoder code. It's the best
 * sort of example I've seen yet on how to use the library.
 *
 * The detection process is rather aggressive. Any hint of garbage
 * that's not ID3v2 at the beginning of the file before the first
 * frame sync signal will make the file be unrecognisable as MPEG.
 * I suspect a FhG VBRI header will make it spit too, since it will
 * make MAD return a "lost sync" error code even though a valid null
 * frame preceeds it, so I shall have to check it and verify.
 *                     -- Jonathon
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "internal.h"

#if (WITH_MPEG == 0)
static jfauderr init(void) { return jfauderr_disabled; }
static jfauderr uninit(void) { return jfauderr_disabled; }
static jfauderr test(File *file UNUSED) { return jfauderr_disabled; }
static jfauderr unload(File *file UNUSED, char *noclose UNUSED) { return jfauderr_disabled; }
static jfauderr getmetainfo(File *file UNUSED, char what UNUSED, char *buf UNUSED, int len UNUSED) { return jfauderr_disabled; }
static jfauderr readsamples(File *file UNUSED, int n UNUSED, PcmBuffer *buf UNUSED, char loop UNUSED) { return jfauderr_disabled; }
static jfauderr readallsamples(File *file UNUSED, PcmBuffer *buf UNUSED) { return jfauderr_disabled; }
static jfauderr getplaytime(File *file UNUSED, float *len UNUSED) { return jfauderr_disabled; }
#else

#include "mad.h"

#ifdef _MSC_VER
#pragma pack(1)
#endif

struct PACKED id3v2head {
	char id3[3];
	unsigned char ver[2];
	unsigned char flags;
	unsigned char size[4];
};
struct PACKED id3v1tag {
	unsigned char tag[3];
	unsigned char title[30];
	unsigned char artist[30];
	unsigned char album[30];
	unsigned char year[4];
	union {
		struct {
			unsigned char comment[30];
		} v1;
		struct {
			unsigned char comment[28];
			unsigned char zero;
			unsigned char tracknum;
		} v1p1;
	} v;
	unsigned char genre;
};

#ifdef _MSC_VER
#pragma pack()
#endif

#define TAG_MAGIC  "TAG"
#define ID3_MAGIC  "ID3"
#define XING_MAGIC "Xing"
#define INFO_MAGIC "Info"

#define XING_NFRAMES  (1)
#define XING_NBYTES   (2)
#define XING_TOC      (4)
#define XING_VBRSCALE (8)
struct xingtag {
	int flags;
	unsigned int nframes;
	unsigned int nbytes;
	unsigned char toc[100];
	unsigned int vbrscale;
};

#define INBUFSZ 20160	// ((448000*(1152/32000))/8)*10
typedef struct {
	WaveformFile_defn

	struct mad_stream stream;
	struct mad_frame  frame;
	struct mad_synth  synth;

	unsigned samplerate,channels;

	unsigned int audiostart;

	unsigned char inbuf[INBUFSZ+MAD_BUFFER_GUARD];
	unsigned int inlen, eof;
        
	unsigned sampsused;
} Mpeg;


#if (DYNAMIC_MPEG == 1)

#ifdef _WIN32
#ifdef PACKED
#undef PACKED
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "libmad.dll"
#else

#include <dlfcn.h>
#define DLLNAME "libmad.so"
#endif

static struct {
#ifdef _WIN32
	HMODULE dll;
#else
	void *dll;
#endif
	char const *mad_version, *mad_copyright;

	void (*mad_stream_init)(struct mad_stream *);
	void (*mad_stream_finish)(struct mad_stream *);
	void (*mad_stream_buffer)(struct mad_stream *, unsigned char const *, unsigned long);
	void (*mad_stream_skip)(struct mad_stream *, unsigned long);
	int  (*mad_stream_sync)(struct mad_stream *);
	char const *(*mad_stream_errorstr)(struct mad_stream const *);
	void (*mad_frame_init)(struct mad_frame *);
	void (*mad_frame_finish)(struct mad_frame *);
	int  (*mad_frame_decode)(struct mad_frame *, struct mad_stream *);
	void (*mad_synth_init)(struct mad_synth *);
	void (*mad_synth_frame)(struct mad_synth *, struct mad_frame const *);
	unsigned long (*mad_bit_read)(struct mad_bitptr *, unsigned int);
} libmad;

#define mad_version libmad.mad_version
#define mad_copyright libmad.mad_copyright
#define mad_stream_init libmad.mad_stream_init
#define mad_stream_buffer libmad.mad_stream_buffer
#define mad_stream_sync libmad.mad_stream_sync
#define mad_stream_skip libmad.mad_stream_skip
#define mad_stream_errorstr libmad.mad_stream_errorstr
#define mad_stream_finish libmad.mad_stream_finish
#define mad_frame_init libmad.mad_frame_init
#define mad_frame_decode libmad.mad_frame_decode
#define mad_frame_finish libmad.mad_frame_finish
#define mad_synth_init libmad.mad_synth_init
#define mad_synth_frame libmad.mad_synth_frame
#define mad_bit_read libmad.mad_bit_read

#define testifloaded() { if (libmad.dll == NULL) return jfauderr_notinited; }
#else

#define testifloaded() { }
#endif


// SoundFileReader functions

static jfauderr init(void)
{
#if (DYNAMIC_MPEG == 1)
        int er=0;

#ifdef _WIN32
        void *sym;
        libmad.dll = LoadLibrary(DLLNAME);

#define gpanc(d,x) GetProcAddress(d,x)
#define gpa(d,x) sym = GetProcAddress(d,x); { \
                if (!sym) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#else
        libmad.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
        dlerror();

#define gpanc(d,x) dlsym(d,x)
#define gpa(d,x) dlsym(d,x); { \
                const char *s; \
                if ((s = dlerror())) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#endif

        if (!libmad.dll) return jfauderr_disabled;
        
	mad_version         = (char const *)gpanc(libmad.dll, "mad_version");
	mad_copyright       = (char const *)gpanc(libmad.dll, "mad_copyright");
        mad_stream_init     = (void*)gpa(libmad.dll, "mad_stream_init");
        mad_stream_buffer   = (void*)gpa(libmad.dll, "mad_stream_buffer");
        mad_stream_sync     = (void*)gpa(libmad.dll, "mad_stream_sync");
        mad_stream_skip     = (void*)gpa(libmad.dll, "mad_stream_skip");
        mad_stream_errorstr = (void*)gpa(libmad.dll, "mad_stream_errorstr");
        mad_stream_finish   = (void*)gpa(libmad.dll, "mad_stream_finish");
        mad_frame_init      = (void*)gpa(libmad.dll, "mad_frame_init");
        mad_frame_decode    = (void*)gpa(libmad.dll, "mad_frame_decode");
        mad_frame_finish    = (void*)gpa(libmad.dll, "mad_frame_finish");
        mad_synth_init      = (void*)gpa(libmad.dll, "mad_synth_init");
	mad_synth_frame     = (void*)gpa(libmad.dll, "mad_synth_frame");
	mad_bit_read        = (void*)gpa(libmad.dll, "mad_bit_read");

        if (er) {
#ifdef _WIN32
                FreeLibrary(libmad.dll);
#else
                dlclose(libmad.dll);
#endif

                libmad.dll = NULL;
                debuglog(DEBUGLVL_GENERAL,"Failed fetching required symbols from " DLLNAME ".");
                return jfauderr_initfailed;
        }

	if (mad_version && mad_copyright) {
		logmsg(jsprintf("    " DLLNAME ": %s\n"
				"    %s",mad_version,mad_copyright));
	}
#endif

	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_MPEG == 1)
        if (libmad.dll) {
#ifdef _WIN32
                FreeLibrary(libmad.dll);
#else
                dlclose(libmad.dll);
#endif

                libmad.dll = NULL;
        }
#endif

	return jfauderr_noerror;
}

static int readxing(struct xingtag *xing, struct mad_bitptr anc_ptr, unsigned int anc_bitlen)
{
	int i;
	unsigned int magic;

	// Xing tag is at least 8 bytes long (header + flags)
	if (anc_bitlen < 8*8) goto problem;

	magic = (unsigned int)mad_bit_read(&anc_ptr, 32);
	if (magic != ExtractBigLong(XING_MAGIC) && magic != ExtractBigLong(INFO_MAGIC)) goto problem;
	xing->flags = (unsigned int)mad_bit_read(&anc_ptr, 32);
	anc_bitlen -= 8*8;

	if (xing->flags & XING_NFRAMES) {
		if (anc_bitlen < 8*4) goto problem;
		xing->nframes = (unsigned int)mad_bit_read(&anc_ptr, 32);
		anc_bitlen -= 8*4;
	}
	if (xing->flags & XING_NBYTES) {
		if (anc_bitlen < 8*4) goto problem;
		xing->nbytes = (unsigned int)mad_bit_read(&anc_ptr, 32);
		anc_bitlen -= 8*4;
	}
	if (xing->flags & XING_TOC) {
		if (anc_bitlen < 8*100) goto problem;
		for (i=0;i<100;i++) {
			xing->toc[i] = (unsigned char)mad_bit_read(&anc_ptr, 8);
		}
		anc_bitlen -= 8*100;
	}
	if (xing->flags & XING_VBRSCALE) {
		if (anc_bitlen < 8*4) goto problem;
		xing->vbrscale = (unsigned int)mad_bit_read(&anc_ptr, 32);
		anc_bitlen -= 8*4;
	}

	// here's where you'd find a LAME tag and whatnot...

	return 0;
problem:
	xing->flags = 0;
	return -1;
}

static int checktag(struct mad_stream const *stream)
{
	struct id3v2head *id3v2 = (struct id3v2head *)stream->this_frame;
	struct id3v1tag *id3v1  = (struct id3v1tag *)stream->this_frame;
	int len = (int)(stream->bufend - stream->this_frame);

	if (len < 0) return -1;

	if ((unsigned)len >= sizeof(struct id3v1tag)) {
		if (!memcmp(id3v1->tag, TAG_MAGIC, 3)) return sizeof(struct id3v1tag);
	}

	if ((unsigned)len >= sizeof(struct id3v2head)) {
		int taglen = 0;

		if (!memcmp(id3v2->id3, ID3_MAGIC, 3) && id3v2->ver[0] < 255 && id3v2->ver[1] < 255 &&
		    id3v2->size[0] < 128 && id3v2->size[1] < 128 && id3v2->size[2] < 128 && id3v2->size[3] < 128) {
			taglen  = (unsigned int)(id3v2->size[0] << 21);
			taglen |= (unsigned int)(id3v2->size[1] << 14);
			taglen |= (unsigned int)(id3v2->size[2] << 7);
			taglen |= (unsigned int)(id3v2->size[3]);
			if (id3v2->flags & 0x10) taglen += 10;	// footer
			taglen  += sizeof(struct id3v2head);
			return taglen;
		}
	}

	return 0;	// heavens knows what it is. probably isn't a tag.
}

static inline signed short madscalesamp(mad_fixed_t sample)
{
        // round
	sample += (1L << (MAD_F_FRACBITS - 16));

        // clip
	if (sample >= MAD_F_ONE) sample = MAD_F_ONE - 1;
        else if (sample < -MAD_F_ONE) sample = -MAD_F_ONE;

        // quantize
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static void processsamples(Mpeg *mpeg, signed short **pbuf, int maxlen)
{
	int togo;
	signed short *buf = *pbuf;

	togo = min(maxlen, (int)(mpeg->synth.pcm.length - mpeg->sampsused));
	if (mpeg->channels == 1) {
		for (; togo>0; togo--) {
			*(buf++) = madscalesamp(mpeg->synth.pcm.samples[0][ mpeg->sampsused++ ]);
		}
	} else {
		for (; togo>0; togo--) {
			*(buf++) = madscalesamp(mpeg->synth.pcm.samples[0][ mpeg->sampsused ]);
			*(buf++) = madscalesamp(mpeg->synth.pcm.samples[1][ mpeg->sampsused ]);
			mpeg->sampsused++;
		}
	}

	*pbuf = buf;
}

static jfauderr fillbuf(File *file, Mpeg *mpeg)
{
	int surplus, readbytes;

	surplus = (int)(mpeg->stream.bufend - mpeg->stream.next_frame);
	if (surplus > 0)
		memmove(mpeg->inbuf, &mpeg->inbuf[mpeg->inlen-surplus], surplus);

	readbytes = sysf_read(&file->fh, &mpeg->inbuf[surplus], INBUFSZ - surplus);
	if (readbytes < 0) return jfauderr_ioerror;
	if (readbytes == 0) {
		memset(&mpeg->inbuf[surplus], 0, MAD_BUFFER_GUARD);
		readbytes = MAD_BUFFER_GUARD;
		mpeg->eof = 1;
	} else mpeg->eof = 0;

	mpeg->inlen = surplus + readbytes;
	mad_stream_buffer(&mpeg->stream, mpeg->inbuf, mpeg->inlen);

	return jfauderr_noerror;
}

static jfauderr decodeframe(File *file, Mpeg *mpeg, int detecting)
{
	jfauderr err;
	int taglen;

	if ((mpeg->stream.bufend - mpeg->stream.next_frame) <= MAD_BUFFER_GUARD)
		if ((err = fillbuf(file,mpeg)) != jfauderr_noerror)
			return err;
	while (mad_frame_decode(&mpeg->frame, &mpeg->stream) < 0) {
		if (mpeg->stream.error == MAD_ERROR_BUFLEN) {
			err = fillbuf(file,mpeg);
			if (err != jfauderr_noerror) return err;
			if (mpeg->eof) return jfauderr_noerror;
			continue;
		} else if (MAD_RECOVERABLE(mpeg->stream.error)) {
			taglen = checktag(&mpeg->stream);
			if (taglen == 0) {
				if (!mpeg->eof) {
					if (detecting) return jfauderr_notrecognised;
					debuglog(DEBUGLVL_WORDY,jsprintf("recoverable frame-level error: %s",
							mad_stream_errorstr(&mpeg->stream)));
				}
			} else {
				debuglog(DEBUGLVL_WORDY,jsprintf("skipping %d byte(s) of tag",taglen));
				mad_stream_skip(&mpeg->stream, taglen);
			}
			continue;
		} else {
			if (detecting) return jfauderr_notrecognised;
			debuglog(DEBUGLVL_GENERAL,jsprintf("unrecoverable frame-level error: %s",
					mad_stream_errorstr(&mpeg->stream)));
			return jfauderr_genericerror;
		}
	}

	mad_synth_frame(&mpeg->synth, &mpeg->frame);
	mpeg->sampsused = 0;

	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	Mpeg *mpeg;
	jfauderr err;
	struct xingtag xing;

	testifloaded();

	file->sf = (SoundFile*)calloc(1,sizeof(Mpeg));
	if (!file->sf) return jfauderr_outofmemory;

	mpeg = (Mpeg*)file->sf;
	
	sysf_seek(&file->fh, 0, SEEK_SET);

	mad_stream_init(&mpeg->stream);
	mad_frame_init(&mpeg->frame);
	mad_synth_init(&mpeg->synth);

	err = decodeframe(file, mpeg, 1);
	if (err != jfauderr_noerror) goto failed;

	if (mpeg->eof) {
		err = jfauderr_notrecognised;
		goto failed;
	}

	if (!readxing(&xing, mpeg->stream.anc_ptr, mpeg->stream.anc_bitlen)) {
		debuglog(DEBUGLVL_TRACE,"Found a Xing tag");
		if (xing.flags & XING_NFRAMES) debuglog(DEBUGLVL_TRACE,jsprintf("num frames = %d",xing.nframes));
		if (xing.flags & XING_NBYTES)  debuglog(DEBUGLVL_TRACE,jsprintf("num bytes = %d",xing.nbytes));
	}

	{
		const char *layer = "?", *mode = "?";
		switch (mpeg->frame.header.layer) {
			case MAD_LAYER_I: layer = "I"; break;
			case MAD_LAYER_II: layer = "II"; break;
			case MAD_LAYER_III: layer = "III"; break;
		}
		switch (mpeg->frame.header.mode) {
			case MAD_MODE_SINGLE_CHANNEL: mode = "single-channel"; break;
			case MAD_MODE_DUAL_CHANNEL:   mode = "dual-channel"; break;
			case MAD_MODE_JOINT_STEREO:   mode = "joint-stereo"; break;
			case MAD_MODE_STEREO:         mode = "stereo"; break;
		}
		mpeg->channels   = MAD_NCHANNELS(&mpeg->frame.header);
		mpeg->samplerate = mpeg->frame.header.samplerate;

		debuglog(DEBUGLVL_WORDY,jsprintf("file is a %dHz %s Layer %s",mpeg->samplerate,mode,layer));
	}

	mpeg->type = type_wave;
	mpeg->format = fmt_mpeg;
	
	return jfauderr_noerror;
failed:
	mad_synth_finish(&mpeg->synth);
	mad_frame_finish(&mpeg->frame);
	mad_stream_finish(&mpeg->stream);

	free(file->sf); file->sf = NULL;

	return err;
}

static jfauderr unload(File *file, char *noclose)
{
	Mpeg *mpeg = (Mpeg *)file->sf;

	testifloaded();

	mad_synth_finish(&mpeg->synth);
	mad_frame_finish(&mpeg->frame);
	mad_stream_finish(&mpeg->stream);

	// caller frees file->sf
	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}

// WaveformFileReader functions
static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	Mpeg *mpeg;
	jfauderr err;
	signed short *rp;
	int lastsampsused;

	testifloaded();

	mpeg = (Mpeg *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	err = pcmbuffer_allocate(buf, n, mpeg->samplerate, mpeg->channels, 16);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;

	rp = (signed short *)buf->data;
	do {
		if (mpeg->sampsused >= mpeg->synth.pcm.length) {
			err = decodeframe(file, mpeg, 0);
			if (err == jfauderr_notcompatible) {
				// means we hit eof
				break;
			} else if (err != jfauderr_noerror) {
				return err;
			}

			if (mpeg->eof) {
				if (loop) {
					sysf_seek(&file->fh, 0, SEEK_SET);
					mpeg->eof = 0;
					continue;
				} else {
					pcmbuffer_free(buf);
					return jfauderr_noerror;
				}
			}
		}

		lastsampsused = mpeg->sampsused;
		processsamples(mpeg, &rp, n);
		n -= (mpeg->sampsused-lastsampsused);
		buf->numsamples += mpeg->sampsused-lastsampsused;
	} while (n>0);
	
	return jfauderr_noerror;
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	return jfauderr_notimpl;
}

static jfauderr getplaytime(File *file, float *len)
{
	*len = -1.0;
	return jfauderr_noerror;
}

#endif

WaveformFileReader wf_mpeg = {
	"MPEG Audio",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_mpeg,
	readsamples,
	readallsamples,
	getplaytime
};

